
(*
 * Module:       Query functions for OE/OS graphs
 *
 * Description:  Design/CPN OG Tool
 *
 * Created by:   Lars Kristensen (kris@daimi.aau.dk)   
 *               Jens B. Jorgensen (jbj@daimi.aau.dk)
 * Date:         29/09/95   
 *
 * Design/CPN (tm)  Copyright (c) Meta Software Corp 1989--1995
 *)

(* $Source: /users/cpn/designcpnR/src/current/oga/OGMLfiles/OEProps.sml,v $

$Log: OEProps.sml,v $
Revision 1.2.4.1  1999/01/04 09:16:28  mailund
char vs. string

Revision 1.2  1996/06/27 14:38:00  cpn
Revised all Query functions for v1.0

Revision 1.1  1996/06/27 14:36:36  cpn
Initial revision
 

*)


(* This file contains functions to analyze OE/OS-graphs and their corresponding 
SCC-graphs. The functions correspond to results in Kurt Jensens book on
Coloured Petri Nets, Volume 2: Analysis Methods ([Jensen]). Each analysis 
function is preceded with a comment stating which result in [Jensen] it 
corresponds to *)

(* Basic List Operation *)
fun CPN'overlap [] _ = false
  | CPN'overlap _ [] = false
  | CPN'overlap (CPN'h::CPN't) CPN'l2 = (mem CPN'l2 CPN'h) orelse (CPN'overlap CPN't CPN'l2);

(* Special-purpose auxiliary functions *)
fun CPN'memBE [] CPN'x = false
  | CPN'memBE CPN'l CPN'x 
   = let
      val CPN'tmp = mkst_BE CPN'x;
      fun CPN'LocMem [] = false
        | CPN'LocMem (CPN'h::CPN't)
	  = CPN'tmp = (mkst_BE CPN'h) orelse (CPN'LocMem CPN't)
     in
      CPN'LocMem CPN'l
     end;
fun CPN'overlapBE [] _ = false
  | CPN'overlapBE (CPN'h::CPN't) CPN'x
    = (CPN'memBE CPN'x CPN'h) orelse (CPN'overlapBE CPN't CPN'x);

(* General-purpose auxiliary functions *)
fun CPN'rmem CPN'x CPN'l = mem CPN'l CPN'x;

fun CPN'int_list_to_string CPN'l =
   let
       fun CPN'lts [] = ""
        |  CPN'lts [(CPN'x:int)] = makestring CPN'x
        |  CPN'lts ((CPN'x:int)::(CPN'y:int)::CPN'l) = 
                (makestring CPN'x)^", "^(CPN'lts (CPN'y::CPN'l));
   in
       "["^CPN'lts CPN'l^"]"
   end;
  
(* exception raised when standard query not available for the class *)
exception std_query_not_avail

(* Reachability Properties *)

(* Prop 2.6 OE/OS/OSP *) 
(* ii *)
fun Reachable (CPN'm1,CPN'm2) = ((NodesInPath(CPN'm1,CPN'm2);
				 true)
				 handle NoPathExists => (false))
    
(*iii*)
fun SccReachable (CPN'm1,CPN'm2) = 
let 
   val CPN'p = SccNodesInPath (NodeToScc CPN'm1,NodeToScc CPN'm2);
in
   CPN'p<>[] 
end;       

(*iv*)
fun AllReachable () = SccNoOfNodes () = 1;

(* Prop 3.7 OS/OSP *)
(* vi *)
fun ReachableSym (CPN'm1,CPN'm2,CPN'Sym) = 
    (case (OESet.GetOEOSSpec ()) of
	 OESpec => raise std_query_not_avail
       | _ =>
	     let
		 val CPN'SymNodes = PredAllNodes CPN'Sym
	     in
		 predlist (fn CPN'n => 
			   (Reachable (CPN'm1,CPN'n) 
			    andalso Reachable (CPN'n,CPN'm2))) 
		 CPN'SymNodes
	     end)
	 
(* vii *)
fun SccReachableSym (CPN'm1,CPN'm2,CPN'Sym) =
    (case (OESet.GetOEOSSpec ()) of
	 OESpec => raise std_query_not_avail
       | _ =>
	     let
		 val CPN'SymNodes = PredAllNodes CPN'Sym
	     in
		 predlist (fn CPN'n => 
			   (SccReachable (CPN'm1,CPN'n)
			    andalso SccReachable (CPN'n,CPN'm2)))
		 CPN'SymNodes
	     end)

(* Boundedness Properties *)

(* Prop 2.7 OE/OS/OSP *)
(* ii *)
fun UpperInteger (CPN'pimf)
   = SearchAllNodes
      (fn _=> true,
       CPN'op.ms'size o CPN'pimf,

       0,Int.max);


fun LowerInteger (CPN'pimf)
   = SearchAllNodes
      (fn _=> true,
       CPN'op.ms'size o CPN'pimf,
       (CPN'op.ms'size o CPN'pimf) InitNode,

       Int.min);


(* iii OE/OS/OSP *)
fun UpperMultiSet (CPN'pimf)
   = let
      fun CPN'ub (CPN'ms1,empty) = CPN'ms1
        | CPN'ub (CPN'ms1,CPN'ms2)
         = let
            fun CPN'LocUb empty = CPN'ms2
              | CPN'LocUb ((CPN'coef,CPN'col)!!CPN'ms1) =

                (case (Int.max(CPN'coef-CPN'op.cf(CPN'col,CPN'ms2),0))

                     of 0 => CPN'LocUb CPN'ms1
                   | CPN'coef  => CPN''add(CPN'coef`CPN'col,
					   (CPN'LocUb CPN'ms1)))
           in
            CPN'LocUb CPN'ms1 
           end
     in
       SearchAllNodes(fn _=>true, CPN'pimf, empty, CPN'ub) 
     end;
     
fun LowerMultiSet (CPN'pimf)
   = let
      fun CPN'lb (_,empty) = empty
        | CPN'lb (CPN'ms1,CPN'ms2)
         = let
            fun CPN'LocLb empty = empty
              | CPN'LocLb ((CPN'coef,CPN'col)!!CPN'ms1)

               = (case Int.min(CPN'coef,CPN'op.cf(CPN'col,CPN'ms2))

                  of 0 => CPN'LocLb CPN'ms1
                   | coef => CPN''add(CPN'coef`CPN'col,(CPN'LocLb CPN'ms1)))
           in
            CPN'LocLb CPN'ms1
           end
     in
      SearchAllNodes(fn _=>true, CPN'pimf, (CPN'pimf InitNode) , CPN'lb)
     end;

(* Prop 3.18 OSP *)
(* ii *)
fun BestUpperMultiSet (CPN'pimf,CPN'eqclasses) =
    (case (OESet.GetOEOSSpec ()) of
	 OSPSpec =>
	     let
		 (* 1. argument is M(p) *)
		 (* 2. argument is the equivalence class *)
		 (* The function finds the colour in the equivalence class *)
		 (* with the largest multiplicity in M(p) *)
		 fun CPN'LocMax _ empty = 0
		   | CPN'LocMax empty _ = 0
		   | CPN'LocMax ((_,CPN'col)!!CPN'ms1) CPN'ms2 =

		     Int.max (CPN'op.cf(CPN'col,CPN'ms2),
			      (CPN'LocMax CPN'ms1 CPN'ms2))

		     
		 fun CPN'comb CPN'EqClass (CPN'm,CPN'gmax) = 

		     Int.max (CPN'gmax,CPN'LocMax CPN'm CPN'EqClass)

		     
		 (* For each equivalence class find the maximum coefficient *)
		 val CPN'coeflist = 
		     map 
		     (fn CPN'eq => SearchAllNodes (fn _ => true,CPN'pimf,0,CPN'comb CPN'eq))
		     CPN'eqclasses
		     
		 fun CPN'sum (_,[]) = empty
		   | CPN'sum ([],_) = empty
		   | CPN'sum (CPN'cf::CPN'xcf,CPN'eq::CPN'xeq) =

		     CPN''add (CPN'cf**CPN'eq,(CPN'sum (CPN'xcf,CPN'xeq)))

	     in
		 CPN'sum (CPN'coeflist,CPN'eqclasses)
	     end
       | _ => raise std_query_not_avail)

fun BestLowerMultiSet (CPN'pimf,CPN'eqclasses) =
    (case (OESet.GetOEOSSpec ()) of
	 OSPSpec =>
	     let
		 (* 1. argument is M(p) *)
		 (* 2. argument is the equivalence class *)
		 (* The function finds the colour in the equivalence class *)
		 (* with the smallest multiplcity in M(p) *)
		 
		 fun CPN'LocMin _ empty CPN'gmin = CPN'gmin
		   | CPN'LocMin empty _ CPN'gmin = CPN'gmin
		   | CPN'LocMin ((_,CPN'col)!!CPN'ms1) CPN'ms2 CPN'gmin =

		     Int.min (CPN'op.cf(CPN'col,CPN'ms2),
			  (CPN'LocMin CPN'ms1 CPN'ms2 CPN'gmin))

		     
		 fun CPN'comb CPN'EqClass (CPN'm,CPN'gmin) = 

		     Int.min (CPN'gmin,CPN'LocMin CPN'm CPN'EqClass CPN'gmin)

		     
		 (* For each equivalence class find the maximum coefficient *)
		 val CPN'coeflist = 
		     map 
		     (fn CPN'eq => SearchAllNodes (fn _ => true,CPN'pimf,
						   (CPN'op.cf ((CPN'op.random CPN'eq),(CPN'pimf InitNode))),
						   CPN'comb CPN'eq))
		     CPN'eqclasses
		     
		 fun CPN'sum (_,[]) = empty
		   | CPN'sum ([],_) = empty
		   | CPN'sum (CPN'cf::CPN'xcf,CPN'eq::CPN'xeq) =

		     CPN''add (CPN'cf**CPN'eq,(CPN'sum (CPN'xcf,CPN'xeq)))

	     in
		 CPN'sum (CPN'coeflist,CPN'eqclasses)
	     end
       | _ => raise std_query_not_avail)
	 
(* Home Properties *)

(**** Prop 2.8 OE/OS/OSP ****)
(* i *)
fun HomeSpace CPN'NodeList
  = let
      val CPN'NodeListExp = (if ((!CPN'env.Timedef <> CPN'env.timenotdef) andalso (with_time ()))
			     then 
				 flatten (map EqualsUntimed CPN'NodeList)
			 else CPN'NodeList)
    in
      SearchSccs ([0],
                fn CPN'scc =>(SccTerminal CPN'scc) andalso
		              (not(CPN'rmem CPN'scc 
				   (remdupl (map NodeToScc CPN'NodeListExp)))),
			      1, fn _ => (), true, fn _=> false)
    end

(* ii*)
fun MinimalHomeSpace () = 
    let 
	val CPN'TermSccs = PredAllSccs SccTerminal
    in
	length CPN'TermSccs
    end

(* iii *)
fun HomeMarking CPN'Node
  = (if ((!CPN'env.Timedef <> CPN'env.timenotdef) andalso (with_time ()))
	 then
	     HomeSpace (EqualsUntimed CPN'Node)
     else
	 SearchSccs ([0],
		     fn CPN'scc =>(SccTerminal CPN'scc) andalso 
		     ((NodeToScc CPN'Node) <> CPN'scc),
		     1, fn _ => (), true, fn _=> false));

fun ListHomeMarkings () =
    let
	val CPN'TermSccs = PredAllSccs SccTerminal
    in
	if length CPN'TermSccs = 1
	    then SccToNodes(hd(CPN'TermSccs))
	else
	    []
    end

fun ListHomeScc () =		
    let 
	val CPN'TermSccs = PredAllSccs SccTerminal
    in
	if length CPN'TermSccs = 1
	    then (hd(CPN'TermSccs))
	else
	    0
    end

(* iv *)
fun HomeMarkingExists ()
 = SearchSccs ([0], SccTerminal, 2, fn _ => 1, 0, op +) = 1;

(* v *)
fun InitialHomeMarking () = 
    (if ((!CPN'env.Timedef <> CPN'env.timenotdef) andalso (with_time ()))
	 then
	     HomeSpace (EqualsUntimed InitNode)
     else
	 SccNoOfNodes () = 1);
	
(* Prop 3.9 OS/OSP *)
(* v *)
fun HomeSpaceSym (CPN'NodeList,CPN'sym) = 
    (case (OESet.GetOEOSSpec ()) of
 	 OESpec => raise std_query_not_avail
       | _ =>
	     let 
		 val CPN'TermSccs = PredAllSccs SccTerminal
	     in
		 (HomeSpace CPN'NodeList) andalso 
		 ((length (PredSccs (CPN'TermSccs,(fn CPN'scc => (predlist CPN'sym (SccToNodes CPN'scc))),
				     NoLimit))) = (length CPN'TermSccs))
	     end)

(* vi *)
fun HomeMarkingSym (CPN'Node,CPN'sym) =	
    (case (OESet.GetOEOSSpec ()) of
 	 OESpec => raise std_query_not_avail
       | _ =>
	     (HomeMarking CPN'Node) andalso 
	     (predlist CPN'sym (SccToNodes (NodeToScc CPN'Node))))
	 

(* Liveness Properties *)

(**** Prop 2.9 OE/OS/OSP ****)
(* i *)
fun DeadMarking CPN'Node = OutArcs CPN'Node = [];

fun ListDeadMarkings () = PredAllNodes Terminal;

(* ii *)
fun SccListDeadMarkings () = 
    let 
	fun CPN'TermAndTriv CPN'n = SccTerminal CPN'n andalso SccTrivial CPN'n
    in
	flatten (map SccToNodes (PredAllSccs CPN'TermAndTriv))
    end;

(* iii *)
fun BEsDead (CPN'BindingElementList,CPN'Node)
   = SearchReachableSccs
      (NodeToScc CPN'Node,
       fn CPN'scc => (CPN'overlapBE (map ArcToBE ((SccToArcs CPN'scc)^^
                    (SccOutArcs CPN'scc))) CPN'BindingElementList),
       1, fn _ => (),true, fn _ => false);

(* iv *)
fun BEsLive CPN'BindingElementList
  = SearchSccs ([0],
                fn CPN'scc => (SccTerminal CPN'scc) andalso
                             (not (CPN'overlapBE (map ArcToBE (SccToArcs CPN'scc))
                              CPN'BindingElementList)),
                1, fn _ => (), true, fn _=> false);

(* v *)
fun TIsLive CPN'TransInstList
  = SearchSccs ([0],
                fn CPN'scc =>(SccTerminal CPN'scc) andalso
                             (not (CPN'overlap (map ArcToTI (SccToArcs CPN'scc))
                              CPN'TransInstList)),
                1, fn _ => (), true, fn _=> false);

(* Prop. 3.10 OS/OSP *)
(* vi *)
fun BEsLiveSym (CPN'BindingElementList,CPN'sym) = 
    (case (OESet.GetOEOSSpec ()) of
	 OESpec => raise std_query_not_avail
       | _ =>
	     let
		 val CPN'TermSccs = PredAllSccs SccTerminal;
	     in
		 (length (PredSccs (CPN'TermSccs,
				    (fn CPN'scc => ((CPN'overlapBE (map ArcToBE (SccToArcs CPN'scc))
						     CPN'BindingElementList) andalso 
						    (predlist CPN'sym (SccToNodes CPN'scc)))),NoLimit))) 
		 = (length CPN'TermSccs) 
	     end)

exception more_than_one_scc;

(* vii *)
fun BEsLiveScc CPN'BindingElementList =
    (case (OESet.GetOEOSSpec ()) of
	 OESpec => raise std_query_not_avail
       | _ =>
    if (SccNoOfNodes () = 1)
	then
	    (CPN'overlapBE (map ArcToBE (SccToArcs (NodeToScc InitNode)))
	     CPN'BindingElementList)
    else
	raise more_than_one_scc)

(* Prop 3.19 OSP *)
(* ii *)
fun TIsStrictlyLiveSym (CPN'BindingElementList,CPN'sym) = 
    (case (OESet.GetOEOSSpec ()) of
	 OSPSpec =>
	     let
		 val CPN'TermSccs = PredAllSccs SccTerminal
	     in
		 (length (PredSccs (CPN'TermSccs,
				    (fn CPN'scc => (not (predlist 
							 (fn CPN'x => (not (mem (map ArcToBE (SccToArcs CPN'scc)) CPN'x)))
							 (CPN'BindingElementList)))
				     andalso (predlist CPN'sym (SccToNodes CPN'scc))),NoLimit))) = (length CPN'TermSccs)
	     end
       | _ => raise std_query_not_avail)
	 
(* iv *)
fun TIsStrictlyLive (CPN'TransInst,CPN'sym) =
    (case (OESet.GetOEOSSpec ()) of
	 OSPSpec =>
	     let
		 val CPN'TermSccs = PredAllSccs SccTerminal
	     in
		 (length (PredSccs (CPN'TermSccs,
				    (fn CPN'scc => (mem (map ArcToTI (SccToArcs CPN'scc)) CPN'TransInst)
				     andalso (predlist CPN'sym (SccToNodes CPN'scc))),NoLimit))) = (length CPN'TermSccs)
	     end
       | _ => raise std_query_not_avail)
	 
(* v *)
fun TIsStrictlyLiveScc CPN'TransInst =
    (case (OESet.GetOEOSSpec ()) of
	 OSPSpec =>
	     (if (SccNoOfNodes () = 1)
		  then
		      (length (PredArcs (EntireGraph,(fn CPN'arc => (ArcToTI CPN'arc) = CPN'TransInst),1)) = 1) 
	      else
		  raise more_than_one_scc)
       |  _ => raise std_query_not_avail)
	 
(* Fairness Properties *)

(**** Prop 2.10 OE/OS/OSP ****)

datatype FairnessProperty = Impartial | Fair | Just | No_Fairness

fun CPN'BEsImpartial CPN'BindingElementList
   = SearchSubgraphSccs
      (fn _ => true,
       fn CPN'arc => not (CPN'memBE CPN'BindingElementList (ArcToBE CPN'arc)),
       not o TrivialSubgraphScc,
       1, fn _ => (), true, fn _ => false);

fun CPN'TIsImpartial CPN'TransInstList
   = SearchSubgraphSccs
      (fn _ => true,
       fn CPN'arc => not (mem CPN'TransInstList (ArcToTI CPN'arc)),
       not o TrivialSubgraphScc,
       1, fn _ => (), true, fn _ => false);

fun CPN'BEsFair CPN'BindingElementList
   = SearchSubgraphSccs
      (fn _ => true,
       fn CPN'arc => not (CPN'memBE (CPN'BindingElementList) (ArcToBE CPN'arc)),
       fn (CPN'scc as (CPN'nodes,_,_)) =>
           not ((TrivialSubgraphScc CPN'scc) orelse
                (not (CPN'overlapBE 
		      (map ArcToBE (flatten (map OutArcs CPN'nodes)))
		      CPN'BindingElementList))),
       1, fn _ => (), true, fn _ => false);

fun CPN'TIsFair CPN'TransInstList
   = SearchSubgraphSccs
      (fn _ => true,
       fn CPN'arc => not (mem CPN'TransInstList (ArcToTI CPN'arc)),
       fn (CPN'scc as (CPN'nodes,_,_)) =>
           not ((TrivialSubgraphScc CPN'scc) orelse
                (not (CPN'overlap (map ArcToTI (flatten (map OutArcs CPN'nodes)))
                              CPN'TransInstList))),
       1, fn _ => (), true, fn _ => false);

fun CPN'BEsJust CPN'BindingElementList
   = SearchSubgraphSccs
      (fn _ => true,
       fn CPN'arc => not (CPN'memBE CPN'BindingElementList (ArcToBE CPN'arc)),
       fn (CPN'scc as (CPN'nodes,CPN'arcs,CPN'outarcs)) =>
           (not (TrivialSubgraphScc CPN'scc) andalso
                (SearchSubgraphCycles 
                  (CPN'nodes,
                   fn CPN'node =>
                    (mem CPN'nodes CPN'node) andalso
                    (CPN'overlapBE CPN'BindingElementList 
		     (map ArcToBE (OutArcs CPN'node))),
                   fn CPN'arc => not (CPN'memBE 
				      CPN'BindingElementList(ArcToBE CPN'arc)),
                   fn _ => true,
                   1, fn _ => (), false, fn _ => true))
                ),
       1, fn _ => (), true, fn _ => false);

fun CPN'TIsJust CPN'TransInstList
   = SearchSubgraphSccs
      (fn _ => true,
       fn CPN'arc => not (mem CPN'TransInstList (ArcToTI CPN'arc)),
       fn (CPN'scc as (CPN'nodes,CPN'arcs,CPN'outarcs)) =>
           (not (TrivialSubgraphScc CPN'scc) andalso
                (SearchSubgraphCycles 
                  (CPN'nodes,
                   fn CPN'node =>
                    (mem CPN'nodes CPN'node) andalso
                    (CPN'overlap CPN'TransInstList (map ArcToTI 
						(OutArcs CPN'node))),
                   fn CPN'arc => not (mem CPN'TransInstList (ArcToTI CPN'arc)),
                   fn _ => true,
                   1, fn _ => (), false, fn _ => true))
                ),
       1, fn _ => (), true, fn _ => false);

fun BEsFairness CPN'BEsl =
    (if (CPN'BEsImpartial CPN'BEsl)
	 then Impartial
     else
	 (if (CPN'BEsFair CPN'BEsl)
	      then Fair
	  else
	      (if (CPN'BEsJust CPN'BEsl)
		   then Just
	       else
		   No_Fairness)))

(* Prop 3.20 OSP *)
fun TIsFairness CPN'TIsl =
    (case (OESet.GetOEOSSpec ()) of
	 OSPSpec =>
	     (if (CPN'TIsImpartial CPN'TIsl)
		  then Impartial
	      else
		  (if (CPN'TIsFair CPN'TIsl)
		       then Fair
		   else
		       (if (CPN'TIsJust CPN'TIsl)
			    then Just
			else
			    No_Fairness)))
       | _ => raise std_query_not_avail)
		  
fun CPN'predlist CPN'f [] = []
  | CPN'predlist CPN'f (CPN'x::CPN'xs)
    = (if CPN'f CPN'x
	   then 
	       (CPN'x::(CPN'predlist CPN'f CPN'xs))
       else
	   CPN'predlist CPN'f CPN'xs);

fun ListImpartialTIs ()
    = CPN'predlist (fn CPN'TI => (TIsFairness [CPN'TI]) = Impartial) TI.All;

fun ListFairTIs ()
    = CPN'predlist (fn CPN'TI =>
		    (let 
			 val CPN'fairness = TIsFairness [CPN'TI]
		     in
			 (CPN'fairness = Fair) orelse
			 (CPN'fairness = Impartial)
		     end)) TI.All;

fun ListJustTIs ()
    = CPN'predlist (fn CPN'TI => 
		    (let 
			 val CPN'fairness = TIsFairness [CPN'TI]
		     in
			 (CPN'fairness = Just) orelse
			 (CPN'fairness = Fair) orelse
			 (CPN'fairness = Impartial)
		     end)) TI.All;

(* Misc. *)

fun EntireGraphCalculated ()
   = PredNodes([0],fn CPN'n => not (FullyProcessed CPN'n),1) = [];

