(*
 * Module: 
Defines a functions determining if there exists a permutation between
two given multi-sets. Moreover, defines functions to apply various
representations of permutations to multi-sets, and functions to
find the usable permutations of a given list, i.e., the ones that
map one given multi-set to another.
 *
 * Description:  Design/CPN OE/OS Tool
 *
 * Created by:   Lars Kristensen (kris@daimi.aau.dk)   
 *               Jens Baek Jorgensen (jbj@daimi.aau.dk)
 * Date:         01/02/96   
 *
 * Design/CPN (tm)  Copyright (c) Meta Software Corp 1989--1995
 *)

(* $Source: /users/cpn/designcpnR/src/current/oga/OGMLfiles/OEUtilities/OEApplyAndTest.sml,v $

$Log: OEApplyAndTest.sml,v $
Revision 1.2  1996/11/27 17:25:11  cpn
*** empty log message ***

Revision 1.1  1996/02/01 11:23:54  jbj
Initial revision
 

*)

val rcsid = "$Header: /users/cpn/designcpnR/src/current/oga/OGMLfiles/OEUtilities/OEApplyAndTest.sml,v 1.2 1996/11/27 17:25:11 cpn Exp $";



structure CPN'OEApplyAndTest =
struct

(* Takes two multi-sets as input. Returns true if there exists
a permutation mapping the first to the second. The coefficient
multi-sets are compared. *)
fun perm_exists ms1 ms2 =
     (CPN'OEConvertMuls.ms_to_cfms ms1 == CPN'OEConvertMuls.ms_to_cfms ms2);


(**************** for permutation symmetries *****************)

(* The lookup function expects a value and a canonical representation of a
permutation as input. It returns the result of applying the permutation on
the value. If it is called with a value that does not appear as first
coordinate in any pair of the list, the value itself us returned. *)
fun lookup CPN'e [] = CPN'e
 |  lookup CPN'e ((CPN'e1,CPN'e2)::CPN'r) = 
                    if CPN'e=CPN'e1 then CPN'e2 else lookup CPN'e CPN'r;

(* The apply function applies a canonical representation of a function
on a given multi-set. It performs the linear extension. *)
fun apply_canon CPN'f empty = empty
 |  apply_canon CPN'f ((CPN'cf,CPN'col)!!CPN'ms) =
                           (CPN'cf,lookup CPN'col CPN'f)!!apply_canon CPN'f CPN'ms;

(* The papply function applies a canonical representation of a function
on a given multi-set of pairs. It performs the linear extension. *)     
fun apply_canon_on_pair CPN'f empty = empty
 |  apply_canon_on_pair CPN'f ((CPN'cf,(CPN'col1,CPN'col2))!!CPN'ms) =
            (CPN'cf,(lookup CPN'col1 CPN'f, lookup CPN'col2 CPN'f))!!
             apply_canon_on_pair CPN'f CPN'ms;

(* Takes two multi-sets of pairs and a list of canonically represented
permutations as input. Returns the sublist of permutations that maps
the first multi-set to the second. *)
fun canon_ok_on_pair CPN'ms1 CPN'ms2 [] = []
 |  canon_ok_on_pair CPN'ms1 CPN'ms2 (CPN'f::CPN'fl) = 
           if ((apply_canon_on_pair CPN'f CPN'ms1)==CPN'ms2) 
           then
              CPN'f::(canon_ok_on_pair CPN'ms1 CPN'ms2 CPN'fl)
           else
              canon_ok_on_pair CPN'ms1 CPN'ms2 CPN'fl;

(* Takes a list of pairs of multi-sets and a list of canonically represented
permutations as argument. It returns true if one of the permutations
maps the first component to the second component of each multi-set pair.
It returns false otherwise. The first time it finds a permutation that works
for all pairs in the list, it stops. *)
local
 fun OKOneCanonicalPair CPN'f [] = true
  |  OKOneCanonicalPair CPN'f ((CPN'ms1,CPN'ms2)::CPN'rest) =
           apply_canon_on_pair CPN'f CPN'ms1 == CPN'ms2
           andalso
           OKOneCanonicalPair CPN'f CPN'rest
in
 fun OKCanonicalPair CPN'l [] = false
  |  OKCanonicalPair CPN'l (CPN'f::CPN'fl) = OKOneCanonicalPair CPN'f CPN'l
                                             orelse
                                             OKCanonicalPair CPN'l CPN'fl
end;


(**************** for rotation symmetries *****************)

(* Takes a function, an offset, and a multi-set as input. Returns
the multi-set that is a rotation of the argument, CPN'i "added" 
using the addition functionto each element. *)
fun apply_rot CPN'add CPN'i empty = empty
 |  apply_rot CPN'add CPN'i ((CPN'cf,CPN'col)!!CPN'ms) =
      (CPN'cf,CPN'add(CPN'col,CPN'i))!!(apply_rot CPN'add CPN'i CPN'ms);

(* Takes a function  and two multi-sets and a list of elements
(representing rotations) as input. Returns the sublist of rotations that maps
the first multi-set to the second.*)
fun rot_ok CPN'add CPN'ms1 CPN'ms2 [] = []
 |  rot_ok CPN'add CPN'ms1 CPN'ms2 (CPN'f::CPN'fl) = 
           if ((apply_rot CPN'add CPN'f CPN'ms1)==CPN'ms2) 
           then
              CPN'f::(rot_ok CPN'add CPN'ms1 CPN'ms2 CPN'fl)
           else
              rot_ok CPN'add CPN'ms1 CPN'ms2 CPN'fl;

(* Takes a rotation function, a list of pairs of multi-sets and a list of rotations as argument.
It returns true if one of the rotations maps the first component to the second
component of each multi-set pair. It returns false otherwise. The first time
it finds a rotation that works for all pairs in the list, it stops. *)
local
 fun OKOneRotation CPN'add CPN'f [] = true
  |  OKOneRotation CPN'add CPN'f ((CPN'ms1,CPN'ms2)::CPN'rest) =
           apply_rot CPN'add CPN'f CPN'ms1 == CPN'ms2
           andalso
           OKOneRotation CPN'add CPN'f CPN'rest
in
 fun OKRotation CPN'add CPN'l [] = false
  |  OKRotation CPN'add CPN'l (CPN'f::CPN'fl) = OKOneRotation CPN'add CPN'f CPN'l
                                                orelse
                                                OKRotation CPN'add CPN'l CPN'fl
end;

end (* struct CPN'OEApplyAndTest *)

