(*
 * Module: Contains functions manipulating restriction and component sets.
The two main functions is one that computes the intersection
between two given restriction sets, and one that computes the
intersection between two given component sets. They implement
the algorithms Intersect and IntersectII from our thesis.
 *
 * Description:  Design/CPN OE/OS Graph Tool
 *
 * Created by:   Lars Kristensen (kris@daimi.aau.dk)   
 *               Jens Baek Jorgensen (jbj@daimi.aau.dk)
 * Date:         01/02/96   
 *
 * Design/CPN (tm)  Copyright (c) Meta Software Corp 1989--1995
 *)

(* $Source: /users/cpn/designcpnR/src/current/oga/OGMLfiles/OEUtilities/OEResAndComp.sml,v $

$Log: OEResAndComp.sml,v $
Revision 1.2.4.1  1999/01/04 09:16:31  mailund
char vs. string

Revision 1.2  1996/11/27 17:25:49  cpn
*** empty log message ***

Revision 1.1  1996/02/01 11:31:07  jbj
Initial revision
 

*)

val rcsid = "$Header: /users/cpn/designcpnR/src/current/oga/OGMLfiles/OEUtilities/OEResAndComp.sml,v 1.2.4.1 1999/01/04 09:16:31 mailund Exp $";

structure CPN'OEResAndComp =
struct

(*** FUNCTIONS MANIPULATING COMPONENT AND RESTRICTION SETS ********)

(* Takes a restriction set as input. Returns a list of all elements
appearing in the right component. *)
fun res_to_rsum CPN'rs = flatten (map CPN'OEListBasics.fst CPN'rs);

(* Takes a restriction set as input. Returns a list of all elements
appearing in the left component. *)
fun res_to_lsum CPN'rs = flatten (map CPN'OEListBasics.snd CPN'rs);

(* Takes a restriction and a restriction set as input. Returns the
restriction set with the restriction removed. I.e., has only an effect
if the restriction is contained in the restriction set. *)
fun rem_res CPN'rs CPN'rss = CPN'OEListBasics.remv CPN'rs CPN'rss;

(* Takes a component and a component set as input. Returns the
component set with the component removed. I.e., has only an effect
if the component is contained in the component set. *)
fun rem_comp CPN'cs CPN'css = CPN'OEListBasics.remv CPN'cs CPN'css;

exception ex_r_common;
exception ex_find_res;

local
(* Aux function for functions to follow *)
fun r_common' (CPN'lc,CPN'rc) [] = false
 |  r_common' (CPN'lc,CPN'rc) ((CPN'lc',CPN'rc')::CPN'rss)
     = if CPN'OEListBasics.intsct_lst CPN'lc CPN'lc' <> [] 
       then true
       else r_common' (CPN'lc,CPN'rc) CPN'rss;

(* Aux function for functions to follow: Takes a restriction and a
restriction set as input. Returns a restriction from the restriction set
such that its left component has something in common with the left
component of the given restriction - if such a restriction exists.
Otherwise an exception is raised. *)
fun find_res (CPN'lc,CPN'rc) [] = raise ex_find_res
 |  find_res (CPN'lc,CPN'rc) ((CPN'lc',CPN'rc')::CPN'rss)
     = if CPN'OEListBasics.intsct_lst CPN'lc CPN'lc' <> [] 
       then (CPN'lc',CPN'rc')
       else find_res (CPN'lc,CPN'rc) CPN'rss;
in

(* Takes two restriction sets as input. Returns true if there are two
restrictions, one from each, with something in common in their left
components. Implements the function common from page 74 in our thesis. *)
fun r_common CPN'rss CPN'rss' = 
 if CPN'rss = [] then false
 else if r_common' (hd CPN'rss) CPN'rss' then true
 else r_common (tl CPN'rss) CPN'rss';

(* Takes two restriction sets as input. Returns a pair of restrictions,
one from each, with something in common in their left components - if
such a pair exists. Otherwise an exception is raised. Implements the
function selectcommon from page 74 in our thesis. *)
fun r_select_common CPN'rss CPN'rss' = 
 if CPN'rss = [] then raise ex_r_common
 else if r_common' (hd CPN'rss) CPN'rss' 
 then (hd CPN'rss, find_res (hd CPN'rss) CPN'rss')
 else r_select_common (tl CPN'rss) CPN'rss'

exception ex_r_delete;

(* Takes a restriction and a restriction set as input. Returns the
restriction set with the restriction removed, i.e., where one restriction
is cut short with the given restriction. It is not checked whether the
given restriction is contained in one of the restrictions of the given
restriction set. Implements the function delete
from page 74 in our thesis.*)
fun r_delete (CPN'lc,CPN'rc) CPN'rss = 
let
  val (CPN'lc',CPN'rc') = find_res (CPN'lc,CPN'rc) CPN'rss;
  val (CPN'l11,CPN'l12) = 
     (CPN'OEListBasics.intsct_lst CPN'lc CPN'lc', CPN'OEListBasics.intsct_lst CPN'rc CPN'rc');
  val (CPN'l21,CPN'l22) =
     (CPN'OEListBasics.subtr_lst CPN'lc' CPN'l11, CPN'OEListBasics.subtr_lst CPN'rc' CPN'l12) 
in
  (CPN'l21,CPN'l22)::(rem_res (CPN'lc',CPN'rc') CPN'rss)    
end;

end (*local*);

exception ex_intersect;

local
(* Takes a restriction set as input. Returns the restriction set witout
the garbage element ([],[])(that may arise during intersection). *)
 fun cleanup CPN'rss = CPN'OEListBasics.remv ([],[]) CPN'rss;

(* Takes four restriction sets as input. Checks if they are consistent -
in accordance with the last six lines in the algorithm Intersect from
page 74 in our thesis. *) 
 fun check CPN'rssb CPN'rss'b CPN'rsse CPN'rss'e = 
    (CPN'OEListBasics.intsct_lst (res_to_rsum CPN'rsse) (res_to_rsum CPN'rss'b) = [])
     andalso
    (CPN'OEListBasics.intsct_lst (res_to_rsum CPN'rss'e) (res_to_rsum CPN'rssb) = []);

(* Aux function for intersect. *)
 fun intersect' CPN'rss CPN'rss' CPN'res false = (CPN'rss, CPN'rss', CPN'res)
  |  intersect' CPN'rss CPN'rss' CPN'res true =
      if r_common CPN'rss CPN'rss'   
      then  
       let
        val ((CPN'c11,CPN'c12),(CPN'c21,CPN'c22)) =
                                         r_select_common CPN'rss CPN'rss';    
        val CPN'l' = CPN'OEListBasics.intsct_lst CPN'c11 CPN'c21;
        val CPN'r' = CPN'OEListBasics.intsct_lst CPN'c12 CPN'c22;
       in
        if (length CPN'l' <> length CPN'r') 
        then 
          raise ex_intersect
        else
         intersect' (r_delete (CPN'l',CPN'r') CPN'rss) 
                    (r_delete (CPN'l',CPN'r') CPN'rss') 
                    ((CPN'l',CPN'r')::CPN'res) true
       end
      else intersect' CPN'rss CPN'rss' CPN'res false
in
(* Takes two restriction sets as input. Returns their intersection. 
Implements the algorithm Intersect from page 74 in our thesis. *) 
 fun intersect CPN'rss CPN'rss' =
    let
     val (CPN'rsse, CPN'rss'e, CPN'res) = intersect' CPN'rss CPN'rss' [] true
    in
     if check CPN'rss CPN'rss' CPN'rsse CPN'rss'e
     then
        cleanup(CPN'res^^CPN'rsse^^CPN'rss'e)
     else
        raise ex_intersect
    end
end;

exception ex_c_common;
exception ex_find_comp;

local
(* Aux function for functions to follow *)
fun c_common' CPN'cs [] = false
 |  c_common' CPN'cs (CPN'cs'::CPN'css)
     = if CPN'OEListBasics.intsct_lst CPN'cs CPN'cs' <> [] 
       then true
       else c_common' CPN'cs CPN'css;

(* Aux function for functions to follow: Takes a component and a
component set as input. Returns a component from the component set
such that its left component has something in common with the left
component of the given component - if such a component exists.
Otherwise an exception is raised. *)
fun find_comp CPN'cs [] = raise ex_find_comp
 |  find_comp CPN'cs (CPN'cs'::CPN'css)
     = if CPN'OEListBasics.intsct_lst CPN'cs CPN'cs' <> [] 
       then CPN'cs'
       else find_comp CPN'cs CPN'css;
in

(* Takes two component sets as input. Returns true if there are two
components, one from each, with something in common in their left
components. Implements the function common from page 78 in our thesis. *)
fun c_common CPN'css CPN'css' = 
 if CPN'css = [] then false
 else if c_common' (hd CPN'css) CPN'css' then true
 else c_common (tl CPN'css) CPN'css';

(* Takes two component sets as input. Returns a pair of components,
one from each, with something in common in their left components - if
such a pair exists. Otherwise an exception is raised. Implements the
function selectcommon from page 78 in our thesis. *)
fun c_select_common CPN'css CPN'css' = 
 if CPN'css = [] then raise ex_c_common
 else if c_common' (hd CPN'css) CPN'css' 
 then (hd CPN'css, find_comp (hd CPN'css) CPN'css')
 else c_select_common (tl CPN'css) CPN'css';

exception ex_c_delete;

(* Takes a component and a component set as input. Returns the
component set with the component removed, i.e., where one component
is cut short with the given component. It is not checked whether the
given component is contained in one of the components of the given
component set. Implements the function delete from page 78 in our thesis.*)
fun c_delete CPN'cs CPN'css = 
let
  val CPN'cs' = find_comp CPN'cs CPN'css;
  val CPN'l1 = CPN'OEListBasics.intsct_lst CPN'cs CPN'cs';
  val CPN'l2 = CPN'OEListBasics.subtr_lst CPN'cs' CPN'l1 
in
  CPN'l2::(rem_comp CPN'cs' CPN'css)    
end
end (*local*)

local
(* Takes a component set as input. Returns the component set witout
the garbage element ([],[])(that may arise during intersection). *)
 fun cleanup CPN'css = CPN'OEListBasics.remv [] CPN'css;

(* Aux function for intersectII. *)
 fun intersectII' CPN'css CPN'css' CPN'res false = 
                                 cleanup(CPN'res^^CPN'css^^CPN'css')
  |  intersectII' CPN'css CPN'css' CPN'res true = 
      if c_common CPN'css CPN'css'
      then
       let
         val (CPN'cs,CPN'cs') = c_select_common CPN'css CPN'css';
         val CPN'c' = CPN'OEListBasics.intsct_lst CPN'cs CPN'cs'
       in
         intersectII' (c_delete CPN'c' CPN'css) 
                      (c_delete CPN'c' CPN'css')
                      (CPN'c'::CPN'res) true
       end
      else intersectII' CPN'css CPN'css' CPN'res false
in
(* Takes two component sets as input. Returns their intersection. 
Implements the algorithm IntersectII from page 78 in our thesis. *) 
 fun intersectII CPN'css CPN'css' = intersectII' CPN'css CPN'css' [] true
end;

(* A "brother" to intersectII that takes indexed component sets as
input, and computes the intersection as if it was ordinary component
sets. *)
fun intersectIIi CPN'icss CPN'icss' = intersectII (map CPN'OEListBasics.snd CPN'icss)
                                                  (map CPN'OEListBasics.snd CPN'icss');

(* Extensions of the intersect functions. Instead of taking two
restriction sets as argument, intersectL takes a list of restriction
sets as argument and returns their intersection. *) 

local 
 fun insct (rss1,rss2) = intersect rss1 rss2; 
in

 fun intersectL CPN'rss = List.foldr insct [] CPN'rss

end;  

local 
 fun insct (css1,css2) = intersectII css1 css2; 
in

 fun intersectIIL CPN'css = List.foldr insct [] CPN'css

end;  

end (* struc CPN'OEResAndComp *)

