(* 
 * Module:          Priority Queue
 *
 * Description:     Heap Priority Queue, imperative implementation
 *
 * Created by:      Torben Bisgaard Haagh
 *
 * Date:            Spring 95
 *)

(* creating a priority queue witn integer key and string element:
 * structure IntPQ = MakePQ(type key = int; val priority = Integer.<);
 * val pq: string IntPQ.pq = IntPQ.create()
 *)

functor MakeHeapPQ(type key; val priority: key * key -> bool) : PRIORITYQUEUE = struct

    type key = key;
    val priority = priority;

    type 'a pq = {heap: (key * 'a) Array.array list, size: int} ref

    exception EmptyPQ;
    
    local
	fun lt ((key1,_),(key2,_)) = priority(key1,key2)

	fun div2  x = Bits.rshift (x,1)
	fun div4  x = Bits.rshift (x,2)
	fun div16 x = Bits.rshift (x,4)

	fun mul2  x = Bits.lshift (x,1)
	fun mul4  x = Bits.lshift (x,2)
	fun mul16 x = Bits.lshift (x,4)

	fun mod16 x = Bits.andb (x,15)

	fun nthhead (x::xs,i) = x::nthhead(xs,i-1)
	  | nthhead (_,0) = nil
	  | nthhead (nil,_) = nil

	fun update (xs,i,v) = 
	    Unsafe.Array.update(List.nth(xs,div16 i), mod16 i, v)
	    handle _ => raise CPN'InternalError "MakePQ.update"

	fun sub (x::_,0) = Unsafe.Array.sub(x,0)
	  | sub (xs, i) =
	    Unsafe.Array.sub(List.nth(xs,div16 i), mod16 i)
	    handle _ => raise CPN'InternalError "MakePQ.sub"
	    
	fun decrease (pq as {heap,size}) = let
	    val n = List.length heap
	in
	    if n>1 andalso size<mul4(n) then
		(* only a quater of the heap is used, decrease it *)
		{heap= nthhead(heap,(div4 n)+1), size=size}
	    else pq
	end

	fun increase (pq as {heap,size},base) =
	    if mul16(List.length heap)<=size then
		(* heap to small, increase it *) 
		{heap= List.@(heap,[Array.array(16,base)]), size=size}
	    else pq

	fun father i = Bits.rshift(i-1,1);   (* (i-1) div 2 *)
	fun lson i   = Bits.lshift(i,1)+1;   (* 2*i + 1 *)
	fun rson i   = Bits.lshift(i,1)+2;   (* 2*1 + 2 *)

	fun swap (A,i,j) = let
	    val tmp = sub(A,i)
	in
	    (update(A,i,sub(A,j)); update(A,j,tmp))
	end;

	fun down (heap,size) node =
	    if (lson node) < size-1 then
		(* there is two sons, swap with the smallest son *)
		if lt(sub(heap,lson node),sub(heap,rson node)) then
		    if lt(sub(heap,lson node),sub(heap,node)) then
			(swap(heap,node,lson node);
			 down (heap,size) (lson node))
		    else ()
		else
		    if lt(sub(heap,rson node),sub(heap,node)) then
			(swap(heap,node,rson node);
			 down (heap,size) (rson node))
		    else ()
	    else if (lson node) = size-1 then
		(* there is only a left son *)
		if lt(sub(heap,lson node),sub(heap,node)) then
		    swap(heap,node,lson node)
		else ()
	    else () (* there is no sons *)

	fun up _ 0 = ()
	  | up (heap,size) node = 
	    if lt(sub(heap,node),sub(heap,father node)) then
		(swap(heap,node,father node); 
		 down (heap,size) (father node);
		 up (heap,size) (father node))
	    else ();
    in
	fun is_empty (ref{size, ...}: 'a pq) = size=0;
	    
	fun create () = ref{heap=nil, size=0};
	    
	fun min ((ref{size=0, ...}): 'a pq) = raise EmptyPQ
	  | min ((ref{heap,...}): 'a pq) = sub(heap,0);
		
	fun insert (pq, key, item) = let
	    val {heap,size} = increase(!pq,(key,item))
	in
	    (update(heap,size,(key,item));
	     up (heap,size) size;
	     pq:={heap=heap,size=size+1})
	end

	fun deletemin (ref{size=0,...}) = raise EmptyPQ
	  | deletemin (pq: 'a pq as ref{heap,size}) =
	    sub(heap,0) before
	    (update(heap,0,sub(heap,size-1));
	     down (heap,size-1) 0;
	     pq:=decrease{heap=heap,size=size-1});

	fun deleteto (ref{size=0,...},_) = nil
	  | deleteto (pq: 'a pq as ref{heap,size},until) =
	    if priority(until,#1(sub(heap,0))) then
		nil
	    else
		deletemin(pq)::deleteto(pq,until)

	fun wipe pq = pq:={heap=nil,size=0}

	fun size (ref {size,...}: 'a pq) = size	    
    end
end
