(* 
 * Module:          Basic BB-Tree functionallity
 *
 * Description:     Basic BB-Tree functionallity to be used in
 *		    various implementations.
 *
 * Created by:      Torben Bisgaard Haagh
 *
 * Date:            Spring 95
 *)

(* $Source: /users/cpn/designcpnR/src/current/oga/OGMLfiles/Attic/OGtree.sml,v $

$Log: OGtree.sml,v $
Revision 1.1.2.2  2000/07/21 07:09:44  kris
Added CPN'

Revision 1.1.2.1  2000/07/20 21:36:27  kris
Initial Revision


*)

val rcsid = "$Header: /users/cpn/designcpnR/src/current/oga/OGMLfiles/Attic/OGtree.sml,v 1.1.2.2 2000/07/21 07:09:44 kris Exp $";

structure CPN'Tree = struct 

    datatype 'a tree = 
	TreeNil
      | TreeNode of {value: 'a, 
		     size: int, 
		     left: 'a tree, 
		     right: 'a tree}
	
    exception EmptyTree;

    fun size (TreeNil) = 0
      | size (TreeNode{size,...}) = size

    fun lson (TreeNode{left,...}) = left
      | lson (TreeNil) = raise EmptyTree;

    fun rson (TreeNode{right,...}) = right
      | rson (TreeNil) = raise EmptyTree;

    fun sons (TreeNode{left,right,...}) = (left,right)
      | sons _ = raise EmptyTree;
    
    (* The function new is added to be used as a smart TreeNode 
     * constructor, to ensure that the invariant I1 is not violated
     *)
    fun new(CPN'v,CPN'l,CPN'r) = 
	TreeNode{value=CPN'v,
		 size= 1 + (size CPN'l) + (size CPN'r),
		 left= CPN'l,
		 right= CPN'r}

    (* The following two functions makes a single counter clock wise 
     * rotation from t1 to t2 or a clock wise from t2 to t1
     * t1.  TreeNode(x,_,s1,TreeNode(y,_,s2,s3))
     * t2.  TreeNode(y,_,TreeNode(x,_,s1,s2),s3)
     *)
    fun single_rotate_ccw (CPN'x, CPN's1, TreeNode{value=CPN'y,left=CPN's2,right=CPN's3,...}) =
	new(CPN'y, new(CPN'x,CPN's1,CPN's2), CPN's3)
      | single_rotate_ccw _ = raise Match;
	    
    fun single_rotate_cw (CPN'y, TreeNode{value=CPN'x,left=CPN's1,right=CPN's2,...}, CPN's3) =
	new(CPN'x, CPN's1, new(CPN'y,CPN's2,CPN's3))
      | single_rotate_cw _ = raise Match;

    (* The following two functions makes a double counter clock wise 
     * rotation from t1 to t2 or a clock wise from t2 to t1
     * t1.  TreeNode(x,_,s1,TreeNode(z,_,TreeNode(y,_,s2,s3),s4))
     * t2.  TreeNode(y,_,TreeNode(x,_,s1,s2),TreeNode(z,_,s3,s4))
     *)
    fun double_rotate_ccw (CPN'x, CPN's1, TreeNode{value=CPN'z,left=TreeNode{value=CPN'y,left=CPN's2,right=CPN's3,...},right=CPN's4,...}) =
	new(CPN'y, new(CPN'x,CPN's1,CPN's2), new(CPN'z,CPN's3,CPN's4))
      | double_rotate_ccw _ = raise Match;
	
    fun double_rotate_cw (CPN'z, TreeNode{value=CPN'x,left=CPN's1,right=TreeNode{value=CPN'y,left=CPN's2,right=CPN's3,...},...},CPN's4) =
	new(CPN'y, new(CPN'x,CPN's1,CPN's2), new(CPN'z,CPN's3,CPN's4))
      | double_rotate_cw _ = raise Match;
	
    (* The value ratio is the weight ration, which tells that one subtree
     * should not have more than ratio times the number of elements in
     * the other subtree
     *)
    val CPN'ratio = 5;

    fun singleton CPN'item = TreeNode{value=CPN'item, 
				  size=1, 
				  left=TreeNil, 
				  right=TreeNil}
	    
    (* The function balance is added to be used as a smart TreeNode 
     * constructor, to ensure that the invariant I2 is not violated
     *)
    fun balance (CPN'p as (CPN'v,CPN'l,CPN'r)) =
	let
	    val CPN'ln = size CPN'l;
	    val CPN'rn = size CPN'r;
	in
	    if CPN'ln+CPN'rn < 2 then
		new CPN'p
	    else if CPN'rn > CPN'ratio * CPN'ln then
		let
		    val (CPN'rl,CPN'rr) = sons CPN'r;
		in
		    if size CPN'rl < size CPN'rr then 
			single_rotate_ccw CPN'p
		    else
			double_rotate_ccw CPN'p
		end
	    else if CPN'ln > CPN'ratio * CPN'rn then
		let
		    val (CPN'll,CPN'lr) = sons CPN'l;
		in
		    if size CPN'lr < size CPN'll then 
			single_rotate_cw CPN'p
		    else
			double_rotate_cw CPN'p
		end
	    else
		new CPN'p
	end;

    fun min (TreeNode{value,left=TreeNil,...}) = value
      | min (TreeNode{left,...}) = min left
      | min (TreeNil) = raise EmptyTree;
	
    fun max (TreeNode{value,right=TreeNil,...}) = value
      | max (TreeNode{right,...}) = max right
      | max (TreeNil) = raise EmptyTree;
	
    fun delmin (TreeNode{value,left=TreeNil,right,...}) = (value,right)
      | delmin (TreeNode{value,left,right,...}) = 
	let
	    val (CPN'value',CPN'left') = delmin left
	in
	    (CPN'value',balance(value,CPN'left',right))
	end
      | delmin _ = raise EmptyTree;
	
    fun delmax (TreeNode{value,left,right=TreeNil,...}) = (value,left)
      | delmax (TreeNode{value,left,right,...}) =
	let
	    val (CPN'value',CPN'right') = delmax right
	in
	    (CPN'value',balance(value,left,CPN'right'))
	end
      | delmax _ = raise EmptyTree;
	    
    fun join (TreeNil,right) = right
      | join (left,TreeNil) = left
      | join (left,right) = 
	if size left < size right then
	    let val (CPN'v,CPN'r) = delmin right in balance(CPN'v,left,CPN'r) end 
	else
	    let val (CPN'v,CPN'l) = delmax left in balance(CPN'v,CPN'l,right) end

    fun fold _ TreeNil CPN'base = CPN'base
      | fold CPN'f (TreeNode{value,left,right,...}) CPN'base =
	fold CPN'f left (CPN'f (value, fold CPN'f right CPN'base));

    fun revfold _ TreeNil CPN'base = CPN'base
      | revfold CPN'f (TreeNode{value,left,right,...}) CPN'base =
	revfold CPN'f right (CPN'f (value, revfold CPN'f left CPN'base));

    fun app _ TreeNil = ()
      | app CPN'f (TreeNode{value,left,right,...}) =
	(app CPN'f left; CPN'f value; app CPN'f right)

    fun revapp _ TreeNil = ()
      | revapp CPN'f (TreeNode{value,left,right,...}) =
	(revapp CPN'f right; CPN'f value; revapp CPN'f left)

end;




