(*
 * Module:       Design/CPN Performance Analysis Library
 *
 * Description:  Create a performance page with an aux. box.
 *               Ensure that there is only one aux. box on
 *               the performance page
 *
 * Created by:   Bo Lindstr<F8>m 
 *               Lisa Wells   
 *
 * Date:         October 1998   
 *
 * Design/CPN    Copyright (c) University of Aarhus 1998
 *)

structure Perf'Page = struct 
val PerfPageName = "Performance_Page";

(* Create box for performance functions - Returns id of the box. *)
fun createPerfBox(CPN'pageid) = 
    let 
	val {h=pgh,name=name,num=num,vis=vis,w=pgw} = 
	    DSRdAttr_GetPageAttr CPN'pageid
	val boxId=
	    DSStr_CreateNode {page=CPN'pageid, x=(~(pgw div 4)),y=0,
			      w=(pgw div 3),h=(pgh div 3),shape=RECTANGLE};
	val setAttr = DSText_SetAttr{obj=boxId,font = Geneva, 
				     size = 12, style = PlainText, 
				     just=LeftJustification}
	val setText = DSText_Put {obj=boxId, 
                             text="(* Performance Functions *)"};
    in
	boxId
    end;

fun createPerfPage (CPN'newName) =
    let
	fun createCorePerfPage (CPN'newName) = 
	    let 
		val pageId = DSStr_NewPage RECTANGLE;
		fun SetPageName ({h=h,name=name,num=num,vis=vis,w=w}, 
				 CPN'newName, pageid) = 
		    DSWtAttr_PageInfo{page=pageid,name=CPN'newName,
				      num=num,vis=false,h=h, w=w};
	    in
		(MakeCpnPage pageId;
		 SetPageName(DSRdAttr_GetPageAttr pageId, 
			     CPN'newName, pageId);
		 pageId)
	    end;

	val pageId  = createCorePerfPage (CPN'newName);
    in
	(createPerfBox(pageId);
	 DSUI_MakePageVisible {page=pageId, front= true};
	 pageId)
    end;

(* Search all pages for performance page.
 * If not found then it is created.     *)
fun getPerfPageId (CPN'whichPerfPage:string) =
    let 
	fun error_PerfPageNotFound() =  
	    DSUI_UserAckMessage("PERFORMANCE WARNING: \n\nPerformance page not found.\nA new one will be created.");
	    
	fun getId [] = 
	    if CPN'whichPerfPage=PerfPageName
		then (error_PerfPageNotFound();
		      createPerfPage(CPN'whichPerfPage))
	      else createPerfPage(CPN'whichPerfPage)
	  | getId (curPage::pageList) =  
	    let val pageAttr = DSRdAttr_GetPageAttr curPage;
		fun getPageName {name=name, num=num, w=w, h=h, vis=vis} = name;
	    in
		if (getPageName (DSRdAttr_GetPageAttr curPage)) = CPN'whichPerfPage
		    then curPage
		else 
		    getId pageList
	    end;
    in
	getId (DSStr_GetPageList ())
    end;    

fun isAuxBox nodeId = (DSRdAttr_GetType nodeId) = 1; (* SYMBOLIC ??? *)

fun noOfAuxBoxesML () = 
    let
        val pageId = getPerfPageId(PerfPageName)
	val nodeList = DSStr_GetNodeList pageId
	fun countBoxesOnPerfPage [] = 0
	  | countBoxesOnPerfPage (curNode::rest) = 
	    if (isAuxBox curNode)
		then 1+(countBoxesOnPerfPage rest)
	    else (countBoxesOnPerfPage rest)
    in
	countBoxesOnPerfPage nodeList
    end;

fun noOfAuxBoxes () = 
	CPN'Edit.out_val(makestring(noOfAuxBoxesML()));
    
fun getPerfBox (CPN'whichPerfPage:string) = 
    let
	val pageId = getPerfPageId(CPN'whichPerfPage)
	val nodeList = DSStr_GetNodeList pageId
	fun findBox [] = createPerfBox(pageId)
	  | findBox (curNode::rest) = 
	    if (isAuxBox curNode)
		then curNode
	    else findBox rest
    in
	findBox nodeList
    end;
   
 end; (* structure *)
